package com.gigaspaces.ucs.service;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.URL;
import java.net.URLConnection;
import java.util.Arrays;

import com.gigaspaces.ucs.domain.in.EventSubscribe;

/*********
 * The Event Receiver is a 'long-polling' client of the UCS manager. It is
 * responsible for sending an http request to the UCS manager and then parses
 * the stream of XML documents sent back from the server in chunks (i.e.
 * 'chunked encoding').
 * 
 * @author barakme
 * 
 */
public class EventRecieverRunnable implements Runnable {

	// logger
	private static java.util.logging.Logger logger = java.util.logging.Logger
	.getLogger(EventRecieverRunnable.class.getName());

	// Communivations constants
	private static final int EVENT_CHANNEL_MAX_CHUNK_SIZE = 50000;
	private static final int EVENT_CHANNEL_READ_TIMEOUT = 60000;
	private static final int EVENT_CHANNEL_CONNECT_TIMEOUT = 5000;

	// fields
	private final Service service;
	private final UCSServiceEventHandler handler;

	private volatile boolean finished = false;

	/********
	 * Launch the event receiver threads. Starts a new thread with the 
	 * provided Runnable.
	 * @param service The UCS Service.
	 * @param handler The events handler. 
	 */
	public static void launchRecieverThread(final Service service,
			final UCSServiceEventHandler handler) {
		new Thread(new EventRecieverRunnable(service, handler)).start();
	}

	/*****
	 * Constructor.
	 * @param service the UCS service.
	 * @param handler the events handler.
	 */
	public EventRecieverRunnable(final Service service,
			final UCSServiceEventHandler handler) {
		this.service = service;
		this.handler = handler;
	}

	private int parseXMLSize(final String line) throws UCSServiceException {
		int size = 0;
		try {
			size = Integer.parseInt(line);
		} catch (final NumberFormatException nfe) {
			throw new UCSServiceException("failed to read size of next chunk: "
					+ line, null);
		}

		if (size > EVENT_CHANNEL_MAX_CHUNK_SIZE) {
			throw new UCSServiceException("Event channel chunk size (" + size
					+ ") exceeds max allowed(" + EVENT_CHANNEL_MAX_CHUNK_SIZE
					+ ")");
		}
		return size;
	}

	@Override
	public void run() {

		try {
			subscribeToEvents(handler);
		} catch (final UCSServiceException e) {
			logger
					.warning("Error while processing events from UCS Manager. Error was: "
							+ e.getMessage());
			logger.warning(Arrays.toString(e.getStackTrace()));

		} catch (final IOException e) {
			logger
					.warning("Communication Error while processing events from UCS Manager. Error was: "
							+ e.getMessage());
			logger.warning(Arrays.toString(e.getStackTrace()));
			
		} finally {
			handler.channelClosed();	
		}

	}

	/******
	 * Stop this receiver from running. Once the next event is received and processed,
	 * the communications channel will be shut down, and the channel closed callback
	 * will be called on the handler. 
	 */
	public void shutdown() {
		this.finished = true;
	}

	/**************
	 * The main logic of the runnable. Sends an HTTP request to the server, and then
	 * reads a stream of XML documents from the response, unmarshals them and sends
	 * the result to the handler.
	 * @param handler the events handler.
	 * @throws UCSServiceException In case of a parsing error.
	 * @throws IOException In case of an IO error.
	 */
	public void subscribeToEvents(final UCSServiceEventHandler handler)
			throws UCSServiceException, IOException {
		final EventSubscribe request = new EventSubscribe();
		request.setCookie(service.getCookie());

		// First marshal the request object using the JAXB Annotations
		final String content = service.getStringFromXMLRequest(request);

		OutputStreamWriter wr = null;
		BufferedReader rd = null;

		try {

			// Send data
			final URL url = service.getURI().toURL();
			final URLConnection conn = url.openConnection();
			conn.setDoOutput(true);
			conn.setConnectTimeout(EVENT_CHANNEL_CONNECT_TIMEOUT);
			conn.setReadTimeout(EVENT_CHANNEL_READ_TIMEOUT);
			wr = new OutputStreamWriter(conn.getOutputStream());
			wr.write(content);
			wr.flush();
			// Get the response

			rd = new BufferedReader(
					new InputStreamReader(conn.getInputStream()));

			String line = rd.readLine();
			while ((line != null) && (!finished)) {
				// get the xml size
				final int size = parseXMLSize(line);

				// read the XML
				final char[] buff = new char[size];
				final int numRead = rd.read(buff);
				if (numRead != size) {
					throw new IOException("Read " + numRead
							+ " bytes from stream, but was expecting " + size);

				}
				final String xml = new String(buff);

				// unmarshal
				final Object event = service.unmarshalEvent(xml);

				// handler performs actual processing
				handler.eventRecieved(event);

				line = rd.readLine();

			}
		} finally {
			if (wr != null) {
				try {
					wr.close();
				} catch (final IOException e) {
					// ignore
				}
			}
			if (rd != null) {
				try {
					rd.close();
				} catch (final IOException e) {
					// ignore
				}
			}

		}

	}

}
