package com.gigaspaces.rackspace;

import java.util.concurrent.Callable;
import java.util.logging.Logger;

import org.jclouds.rackspace.cloudservers.CloudServersClient;
import org.jclouds.rackspace.cloudservers.domain.Server;
import org.jclouds.rackspace.cloudservers.domain.ServerStatus;
import org.jclouds.rackspace.cloudservers.options.CreateServerOptions;

public abstract class AbstractStartServerTask implements Callable<Server> {
	private static final int DEFAULT_RETRIES = 3;
	private static final int WAIT_THREAD_SLEEP_MILLIS = 10000;
	private static final int WAIT_TIMEOUT_MILLIS = 360000;
	private static final Logger logger = 
		Logger.getLogger(AbstractStartServerTask.class.getName());
	protected CloudServersClient rackspaceClient;
	protected int retries = DEFAULT_RETRIES;

	protected AbstractStartServerTask(final CloudServersClient rackspaceClient) {
		this.rackspaceClient = rackspaceClient;
	}

	public Server createServer(final String serverName, final int flavorId, final int imageId,
			final CreateServerOptions options) {
		logger.info("Loading Server - Name:" + serverName + " FlavorId:" + flavorId + " ImageId:"
				+ imageId + "... ");
		final Server server = rackspaceClient.createServer(serverName, imageId, flavorId, options);
		return server;
	}

	/*********
	 * Periodically gets the server status from Rackspace, until the server's
	 * status changes to ACTIVE, or a timeout expires.
	 * 
	 * @param serverId
	 *            The server ID.
	 * @return The server status - should always be ACTIVE.
	 */
	protected ServerStatus waitUntilServerIsActive(final int serverId) {
		final long endTime = System.currentTimeMillis() + WAIT_TIMEOUT_MILLIS; // 10min
		while (true) {
			Server server = rackspaceClient.getServer(serverId);
			if (server.getStatus().compareTo(ServerStatus.ACTIVE) == 0) {
				return server.getStatus();
			}
			
			if (System.currentTimeMillis() > endTime) {
				throw new IllegalStateException("Server [ " + serverId
						+ " ] has been starting up for more more than 6 minutes!");
			}
			
			logger.fine("Server Status (" + server.getStatus().value()
					+ ") still not active, please wait...");
			try {
				Thread.sleep(WAIT_THREAD_SLEEP_MILLIS);
			} catch (InterruptedException e) {
				// ignore.
			}

		}
		
	}

}
