package com.gigaspaces.rackspace;

import java.io.IOException;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.annotation.PreDestroy;

import org.jclouds.compute.ComputeServiceContext;
import org.jclouds.compute.ComputeServiceContextFactory;
import org.jclouds.rackspace.cloudservers.CloudServersClient;
import org.jclouds.rackspace.cloudservers.domain.Server;
import org.jclouds.rackspace.cloudservers.options.ListOptions;
import org.openspaces.admin.Admin;

import com.google.inject.Module;

public class Deployer {
	private static final int ESM_STARTUP_TIMEOUT = 5;
	private static final int IMAGE_CENTOS_55 = 51;
	private static final int FLAVOR_2MB = 5;

	private static final Logger logger = Logger.getLogger(Deployer.class.getName());

	private final ComputeServiceContext context;
	private final CloudServersClient rackspaceClient;
	private final ExecutorService exe;

	private final int deploymentPoolSize = 4;
	private int imageId = IMAGE_CENTOS_55;
	private int flavorId = FLAVOR_2MB;

	/***
	 * Constructor.
	 * 
	 * @param user
	 *            Rackspace username.
	 * @param apiKey
	 *            Rackspace API Access Key.
	 * @throws IOException
	 *             if there was a problem while initializing the Rackspace
	 *             context.
	 */
	public Deployer(final String user, final String apiKey) throws IOException {
		final Properties overrides = new Properties();

		final Set<Module> wiring = new HashSet<Module>();
		this.context = new ComputeServiceContextFactory().createContext("cloudservers", user,
				apiKey, wiring, overrides);

		this.rackspaceClient = CloudServersClient.class.cast(context.getProviderSpecificContext()
				.getApi());
		exe = Executors.newFixedThreadPool(deploymentPoolSize);
		/*
		 * Template template =
		 * context.getComputeService().templateBuilder().build();
		 * template.getOptions().runScript(newStringPayload(script));
		 * context.getComputeService().runNodesWithTag(tag, 2, template);
		 */
	}

	@PreDestroy
	public void destroy() {
		context.close();
		exe.shutdown();
	}

	public List<Server> getAllServers() {
		return this.rackspaceClient.listServers(new ListOptions().withDetails());
	}

	public Server getServerWithIP(final String ip) {
		final List<Server> servers = getAllServers();
		for (final Server server : servers) {
			// check for nulls everywhere!
			if ((server != null) && (server.getAddresses() != null)) {

				if ((server.getAddresses().getPrivateAddresses() != null)
						&& (server.getAddresses().getPrivateAddresses().contains(ip))) {
					return server;
				}
				if ((server.getAddresses().getPublicAddresses() != null)
						&& (server.getAddresses().getPublicAddresses().contains(ip))) {
					return server;
				}
			}
		}

		return null;
	}

	public Server getServer(final String name) {
		List<Server> servers = getAllServers();
		for (Server server : servers) {
			if ((server.getName() != null) && server.getName().equals(name)) {
				return server;
			}
		}
		return null;

	}

	public Server launchESMMachine(final String name, final Admin admin) {

		final Future<Server> future = launchESMMachineAsync(name, admin);
		Server server = null;
		try {
			server = future.get(ESM_STARTUP_TIMEOUT, TimeUnit.MINUTES);
		} catch (final InterruptedException e) {
			logger.log(Level.SEVERE, "Failed to start ESM", e);
		} catch (final ExecutionException e) {
			logger.log(Level.SEVERE, "Failed to start ESM", e);
		} catch (final TimeoutException e) {
			logger.log(Level.SEVERE, "Starting the ESM is taking too long!", e);
		}
		return server;
	}

	public Future<Server> launchESMMachineAsync(final String name, final Admin admin) {

		final Future<Server> future = exe.submit(new StartESMTask(rackspaceClient, name, flavorId,
				imageId, null, admin));
		return future;
	}

	public Server launchGSAMachine(final String lusMachineIP, final String serverName,
			final Admin admin) {
		final Future<Server> future = launchGSAMachineAsync(lusMachineIP, serverName, admin);
		Server server = null;
		try {
			server = future.get(ESM_STARTUP_TIMEOUT, TimeUnit.MINUTES);

		} catch (final InterruptedException e) {
			logger.log(Level.SEVERE, "Failed to start ESM", e);
		} catch (final ExecutionException e) {
			logger.log(Level.SEVERE, "Failed to start ESM", e);
		} catch (final TimeoutException e) {
			logger.log(Level.SEVERE, "Starting the ESM is taking too long!", e);
		}
		return server;
	}

	public Future<Server> launchGSAMachineAsync(final String lusMachineIP, final String serverName,
			final Admin admin) {
		final Future<Server> future = exe.submit(new StartESMTask(rackspaceClient, serverName,
				flavorId, imageId, lusMachineIP, admin));
		return future;
	}

	public void setFlavorId(final int flavorId) {
		this.flavorId = flavorId;
	}

	public void setImageId(final int imageId) {
		this.imageId = imageId;
	}

	public boolean shutdownMachine(final int serverId) {
		return rackspaceClient.deleteServer(serverId);
	}

}
