package com.gigaspaces.rackspace;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.taskdefs.optional.ssh.SSHExec;
import org.jclouds.rackspace.cloudservers.CloudServersClient;
import org.jclouds.rackspace.cloudservers.domain.Server;
import org.jclouds.rackspace.cloudservers.options.CreateServerOptions;
import org.openspaces.admin.Admin;
import org.openspaces.admin.machine.Machine;

import com.gigaspaces.rackspace.utils.FilesUtils;

/***********
 * Starts up a Rackspace cloud server and sets up GigaSpaces on it.
 * 
 * @author barakme
 * 
 */
public class StartESMTask extends AbstractStartServerTask {

	private static final int CLUSTER_JOIN_TIMEOUT_MILLIS = 60000;

	private static final int SSH_COMMAND_TIMEOUT = 30000;

	private static final int LUS_CONNECTION_RETRIES = 6;

	private static final int LUS_PORT = 4164;

	private static final int SSH_CONNECTION_RETRIES = 3;

	private static final int SSH_PORT = 22;

	private static final int CONNECTION_TEST_INTERVAL_MILLIS = 5000;

	private static final int CONNECTION_TEST_CONNECT_TIMEOUT_MILLIS = 10000;

	private static Logger logger = Logger.getLogger(StartESMTask.class.getName());

	private final String serverName;
	private final int flavorId;
	private final int imageId;
	private final String lusIP;

	private Admin admin;

	/********
	 * Constructor.
	 * 
	 * @param rackspaceClient
	 *            The Rackspace client from the jclouds context.
	 * @param serverName
	 *            The name of the new server.
	 * @param flavorId
	 *            Rackspace server flavor.
	 * @param imageId
	 *            Rackspace server image.
	 * @param lusIP
	 *            IP of the Lookup service to use, or null if this server should
	 *            also run the LUS.
	 * @param admin
	 *            an active admin API instance, or null if this is the first
	 *            node in the cluster.
	 **/
	public StartESMTask(final CloudServersClient rackspaceClient, final String serverName,
			final int flavorId, final int imageId, final String lusIP, final Admin admin) {
		super(rackspaceClient);
		this.serverName = serverName;
		this.flavorId = flavorId;
		this.imageId = imageId;
		this.lusIP = lusIP;
		this.admin = admin;

	}

	@Override
	public Server call() throws Exception {

		final Server server = createServer();

		// When running inside Rackspace, it is better to use the private IP.
		String targetIP = (String) server.getAddresses().getPrivateAddresses().toArray()[0];
		String lusLocator = null;
		if (lusIP == null) {
			lusLocator = targetIP;
		} else {
			lusLocator = lusIP;
		}

		if (logger.isLoggable(Level.INFO)) {
			logger.info("ESM Server was created.");
			logger.info("Password: " + server.getAdminPass());
			logger.info("Public IP: "
					+ Arrays.toString(server.getAddresses().getPublicAddresses().toArray()));
			logger.info("Private IP: "
					+ Arrays.toString(server.getAddresses().getPrivateAddresses().toArray()));
			logger.info("Target IP for connection: " + targetIP);
			logger.info("LUS IP for Locator: " + lusLocator);
			logger.info("Connect with putty using: putty -pw " + server.getAdminPass() + " root@"
					+ server.getAddresses().getPublicAddresses().toArray()[0]);
		}

		// Server is not active yet, so need to wait.
		logger.info("Please wait until server is active");
		super.waitUntilServerIsActive(server.getId());
		logger.info("Server is now active and can be reached at the IPs specified above");

		// checking for SSH connection
		if (!checkConnection(targetIP, SSH_PORT, SSH_CONNECTION_RETRIES)) {
			logger.severe("Failed to open a connection to the target machine (" + targetIP + ") "
					+ "on port 22. Please check firewall settings");
			return null;
		}

		// SSH is ready, call startup script.
		logger
				.info("Calling startup script on target: " + targetIP + " with LOCATOR="
						+ lusLocator);
		String cmd = null;
		if (this.lusIP == null) {
			cmd = "chmod +x /opt/start-esm.sh;/opt/start-esm.sh " + lusLocator + " &";
		} else {
			cmd = "chmod +x /opt/start-esm.sh;" + "/opt/start-esm.sh " + lusLocator + " agent &";
		}

		sshCommand(targetIP, cmd, "root", server.getAdminPass());

		// wait for the server to join
		if (admin != null) {
			waitForServerToJoin(targetIP);
		}

		// for LUS node, check connection to LUS
		if (lusIP == null) {
			if (!checkConnection(targetIP, LUS_PORT, LUS_CONNECTION_RETRIES)) {
				logger.severe("Failed to open a connection to the target machine (" + targetIP
						+ ") " + "on port 4164. Please check firewall settings");
				return null;
			}
		}

		// At this point, the server is ready to be used.
		return server;
	}

	private Server createServer() throws IOException {
		// Set up the script that should run on the server once it is up.
		final CreateServerOptions options = new CreateServerOptions();

		options.withFile("/opt/start-esm.sh", FilesUtils
				.getBytesFromResource("config/start-esm.sh"));

		// Start the server. This is a blocking call.
		logger.info("Starting ESM Server");
		final Server server = super.createServer(serverName, flavorId, imageId, options);
		return server;
	}

	private void waitForServerToJoin(final String address) {

		long startTime = System.currentTimeMillis();
		while (true) {
			logger.fine("Checking if " + address + " has joined");
			final Machine machine = admin.getMachines().getMachineByHostAddress(address);
			if (machine != null) {
				logger.fine(address + " has joined the cluster!");
				return;
			}
			logger.fine(address + " has not joined yet, sleeping");
			long now = System.currentTimeMillis();
			if (now - startTime > CLUSTER_JOIN_TIMEOUT_MILLIS) {
				logger.severe("Server " + address
						+ " failed to join cluster before timeout expired");
				throw new IllegalStateException("Server " + address
						+ " failed to join cluster before timeout expired");
			}
		}

	}

	private boolean checkConnection(final String ip, final int port, final int retries) {
		for (int i = 0; i < retries; ++i) {
			final Socket sock = new Socket();
			try {
				logger.info("Checking connection to: " + ip + ":" + port);
				sock.connect(new InetSocketAddress(ip, port),
						CONNECTION_TEST_CONNECT_TIMEOUT_MILLIS);
				sock.close();
				return true;
			} catch (final IOException e) {
				// ignore
			}

			try {
				Thread.sleep(CONNECTION_TEST_INTERVAL_MILLIS);
			} catch (final InterruptedException e) {
				// ignore
			}
		}

		return false;

	}

	private void sshCommand(final String host, final String command, final String username,
			final String password) {

		final SSHExec task = new SSHExec();
		task.setCommand(command);
		task.setHost(host);
		task.setTrust(true);
		task.setUsername(username);
		task.setPassword(password);
		task.setTimeout(SSH_COMMAND_TIMEOUT);

		try {
			logger.info("Executing command: " + command + " on " + host);
			task.execute();
		} catch (final BuildException e) {
			// There really should be a better way to check that this is
			// a Timeout
			if (e.getMessage() != null && e.getMessage().contains("Timeout")) {
				logger.info("ssh execution ends with timeout: " + e.getCause());
			} else {
				throw new IllegalStateException("Command " + command + " failed to execute: "
						+ e.getMessage(), e);
			}

		}
	}

}
